ConstructionActions = {
    HEIGHT_UPDATE = 0.1
}
local ConstructionActions_mt = Class(ConstructionActions)

function ConstructionActions.new(customMt)
	local self = setmetatable({}, customMt or ConstructionActions_mt)
    self.heightChanger = 0.0
    self.currentHeight = 0.0
    self.shouldShowPastureExtensionHelper = false
    self.alignToNavMesh = false
    self.brush = nil
    self:init()

	return self
end

function ConstructionActions:init()
    ConstructionScreen.setBrush = Utils.appendedFunction(ConstructionScreen.setBrush, self.setBrush)
    ConstructionScreen.registerMenuActionEvents = Utils.overwrittenFunction(ConstructionScreen.registerMenuActionEvents, self.registerActionEventIfNeeded)
    ConstructionScreen.onClose = Utils.appendedFunction(ConstructionScreen.onClose, self.unregisterActionEventIfNeeded)
    ConstructionScreen.updateBrushActionTexts = Utils.overwrittenFunction(ConstructionScreen.updateBrushActionTexts, self.updateBrushActionTexts)
end

function ConstructionActions:updateNavMeshes(visible)
    self:toggleNavMeshInputActions(visible)

    if self.brush ~= nil and self.brush.placeable ~= nil then
        if self.brush.placeable.spec_husbandryAnimals ~= nil and self.brush.placeable.spec_husbandryAnimals.helperNavMeshPlane ~= nil then
            setVisibility(self.brush.placeable.spec_husbandryAnimals.helperNavMeshPlane, visible)
        elseif self.brush.placeable.spec_pastureExtension ~= nil and self.brush.placeable.spec_pastureExtension.helperNavMeshPlane ~= nil then
            setVisibility(self.brush.placeable.spec_pastureExtension.helperNavMeshPlane, visible)
        end
    end

    local allAvailablePastures = g_currentMission.pastureSystem:availablePastures()

    for i,v in ipairs(allAvailablePastures) do
        setVisibility(v.spec_husbandryAnimals.helperNavMeshPlane, visible)
    end

    local allPastureExtensions = g_currentMission.pastureSystem:allAvailablePastureExtensions()

    for extensionIndex, extension in ipairs(allPastureExtensions) do
        setVisibility(extension.spec_pastureExtension.helperNavMeshPlane, visible)
    end
end

function ConstructionActions:toggleNavMeshInputActions(active)
    if self.lowerNavMeshId ~= nil or self.liftNavMeshId ~= nil or self.alignExtensionsNavMeshId ~= nil then
        g_inputBinding:setActionEventActive(self.lowerNavMeshId, active)
        g_inputBinding:setActionEventActive(self.liftNavMeshId, active)
        g_inputBinding:setActionEventActive(self.alignExtensionsNavMeshId, active)
    end
end

function ConstructionActions:lowerNavMesh()
    self.heightChanger = -ConstructionActions.HEIGHT_UPDATE
    self.currentHeight = self.currentHeight - ConstructionActions.HEIGHT_UPDATE
    self:updateNavMeshHeight()
end

function ConstructionActions:liftNavMesh()
    self.heightChanger = ConstructionActions.HEIGHT_UPDATE
    self.currentHeight = self.currentHeight + ConstructionActions.HEIGHT_UPDATE
    self:updateNavMeshHeight()
end

function ConstructionActions:updateNavMeshHeight()
    if self.brush ~= nil and self.brush.placeable ~= nil then
        if self.brush.placeable.spec_husbandryAnimals ~= nil and self.brush.placeable.spec_husbandryAnimals.helperNavMeshPlane ~= nil then
            local x, y, z = getTranslation(self.brush.placeable.spec_husbandryAnimals.helperNavMeshPlane)
            y = y + self.heightChanger
            setTranslation(self.brush.placeable.spec_husbandryAnimals.helperNavMeshPlane, x, y, z)
        elseif self.brush.placeable.spec_pastureExtension ~= nil and self.brush.placeable.spec_pastureExtension.helperNavMeshPlane ~= nil then
            local x, y, z = getTranslation(self.brush.placeable.spec_pastureExtension.helperNavMeshPlane)
            y = y + self.heightChanger
            setTranslation(self.brush.placeable.spec_pastureExtension.helperNavMeshPlane, x, y, z)
        end
    end

    local allAvailablePastures = g_currentMission.pastureSystem:availablePastures()

    for i,v in ipairs(allAvailablePastures) do
        local x, y, z = getTranslation(v.spec_husbandryAnimals.helperNavMeshPlane)
        y = y + self.heightChanger
        setTranslation(v.spec_husbandryAnimals.helperNavMeshPlane, x, y, z)
    end

    local allPastureExtensions = g_currentMission.pastureSystem:allAvailablePastureExtensions()

    for extensionIndex, extension in ipairs(allPastureExtensions) do
        local x, y, z = getTranslation(extension.spec_pastureExtension.helperNavMeshPlane)
        y = y + self.heightChanger
        setTranslation(extension.spec_pastureExtension.helperNavMeshPlane, x, y, z)
    end
end

function ConstructionActions:toggleNavMesh()
    self.shouldShowPastureExtensionHelper = not self.shouldShowPastureExtensionHelper
    self:updateNavMeshes(self.shouldShowPastureExtensionHelper)
end

function ConstructionActions.registerActionEventIfNeeded(self, superFunc, hasMenuButtons)
    superFunc(self, hasMenuButtons)

    if g_ep_constructionActions.toggleNavMeshId ~= nil then
        g_inputBinding:setActionEventActive(g_ep_constructionActions.toggleNavMeshId, true)
        return
    end

    local toggleNavMeshIdValid, toggleNavMeshId = g_inputBinding:registerActionEvent(InputAction.EP_TOGGLE_NAVMESH, g_ep_constructionActions, g_ep_constructionActions.toggleNavMesh, false, true, false, true)
    local setParentPastureValid, setParentPastureId = g_inputBinding:registerActionEvent(InputAction.EP_SET_PARENT_PASTURE, g_ep_constructionActions, g_ep_constructionActions.openPasturePicker, false, true, false, true)

    if toggleNavMeshIdValid and toggleNavMeshId ~= nil then
        g_inputBinding:setActionEventActive(toggleNavMeshId, true)
        g_ep_constructionActions.toggleNavMeshId = toggleNavMeshId
    end

    if setParentPastureValid and setParentPastureId ~= nil then
        g_inputBinding:setActionEventActive(setParentPastureId, false)
        g_ep_constructionActions.setParentPastureId = setParentPastureId
    end

    local lowerNavMeshIdValid, lowerNavMeshId = g_inputBinding:registerActionEvent(InputAction.EP_LOWER_NAVMESH, g_ep_constructionActions, g_ep_constructionActions.lowerNavMesh, false, false, true, true)
    local liftNavMeshIdValid, liftNavMeshId = g_inputBinding:registerActionEvent(InputAction.EP_LIFT_NAVMESH, g_ep_constructionActions, g_ep_constructionActions.liftNavMesh, false, false, true, true)
    local alignExtensionsNavMeshIdValid, alignExtensionsNavMeshId = g_inputBinding:registerActionEvent(InputAction.EP_ALIGN_EXTENSIONS_NAVMESH, g_ep_constructionActions, g_ep_constructionActions.alignExtensionsNavMeshes, false, true, false, true)

    if lowerNavMeshIdValid and liftNavMeshIdValid and alignExtensionsNavMeshIdValid and lowerNavMeshId ~= nil or liftNavMeshId ~= nil and alignExtensionsNavMeshId ~= nil then
        g_inputBinding:setActionEventActive(lowerNavMeshId, false)
        g_ep_constructionActions.lowerNavMeshId = lowerNavMeshId
        g_inputBinding:setActionEventActive(liftNavMeshId, false)
        g_ep_constructionActions.liftNavMeshId = liftNavMeshId
        g_inputBinding:setActionEventActive(alignExtensionsNavMeshId, false)
        g_ep_constructionActions.alignExtensionsNavMeshId = alignExtensionsNavMeshId
    end
end

function ConstructionActions.unregisterActionEventIfNeeded()
    g_ep_constructionActions.shouldShowPastureExtensionHelper = false
    g_ep_constructionActions:updateNavMeshes(false)

    if g_ep_constructionActions.toggleNavMeshId ~= nil then
        g_inputBinding:setActionEventActive(g_ep_constructionActions.toggleNavMeshId, false)
    end

    if g_ep_constructionActions.setParentPastureId ~= nil then
        g_inputBinding:setActionEventActive(g_ep_constructionActions.setParentPastureId, false)
    end

    if g_ep_constructionActions.lowerNavMeshId ~= nil then
        g_inputBinding:setActionEventActive(g_ep_constructionActions.lowerNavMeshId, false)
    end

    if g_ep_constructionActions.liftNavMeshId ~= nil then
        g_inputBinding:setActionEventActive(g_ep_constructionActions.liftNavMeshId, false)
    end

    if g_ep_constructionActions.alignExtensionsNavMeshId ~= nil then
        g_inputBinding:setActionEventActive(g_ep_constructionActions.alignExtensionsNavMeshId, false)
    end
end

function ConstructionActions:setBrush(brush, skipMenuUpdate)
    local showNavMeshHelper = false
    g_ep_constructionActions.brush = brush

    if brush ~= nil and brush.loadingPlaceable ~= nil and brush.loadingPlaceable.specializations ~= nil then
        showNavMeshHelper = SpecializationUtil.hasSpecialization(PlaceablePastureExtension, brush.loadingPlaceable.specializations) or SpecializationUtil.hasSpecialization(PlaceableHusbandryAnimalsExtension, brush.loadingPlaceable.specializations)
    end

    if brush ~= nil and brush.destructMode then
        showNavMeshHelper = true
    end

    if showNavMeshHelper and not g_ep_constructionActions.shouldShowPastureExtensionHelper then
        g_ep_constructionActions:toggleNavMesh()
    elseif not showNavMeshHelper and g_ep_constructionActions.shouldShowPastureExtensionHelper then
        g_ep_constructionActions:toggleNavMesh()
    end
end

function ConstructionActions.updateBrushActionTexts(self, superFunc)
    superFunc(self)

	if self.brush == nil or self.brush.lastPlaceable == nil or
        self.brush.lastPlaceable.isDeleted or
        not SpecializationUtil.hasSpecialization(PlaceablePastureExtension, self.brush.lastPlaceable.specializations) or
        #g_currentMission.pastureSystem:availablePastures(self.brush.lastPlaceable.parentPasture) <= 0
    then
        g_ep_constructionActions.currentPastureExtension = nil
        g_inputBinding:setActionEventActive(g_ep_constructionActions.setParentPastureId, false)
    else
        g_ep_constructionActions.currentPastureExtension = self.brush.lastPlaceable
        g_inputBinding:setActionEventActive(g_ep_constructionActions.setParentPastureId, true)
    end
end

function ConstructionActions:openPasturePicker()
    if self.currentPastureExtension ~= nil then
        g_pasturePickerDialog:setCallback(self.currentPastureExtension.onPastureSelected, self.currentPastureExtension)
        g_pasturePickerDialog:setCurrentParentPasture(self.currentPastureExtension.parentPasture)
        g_gui:showDialog("EP_PasturePickerDialog")
    end
end

function ConstructionActions:alignExtensionsNavMeshes()
    self.heightChanger = 0
    self.currentHeight = 0
    self.alignToNavMesh = not self.alignToNavMesh

    local allAvailablePastures = g_currentMission.pastureSystem:availablePastures()

    for i,v in ipairs(allAvailablePastures) do
        setTranslation(v.spec_husbandryAnimals.helperNavMeshPlane, 0, 0, 0)
    end

    local allPastureExtensions = g_currentMission.pastureSystem:allAvailablePastureExtensions()

    for extensionIndex, extension in ipairs(allPastureExtensions) do
        local navMeshYPosition = extension.spec_pastureExtension.navMeshYPosition
        if navMeshYPosition ~= nil then
            if self.alignToNavMesh then
                local _, alignedHeight, _ = worldToLocal(extension.spec_pastureExtension.navMeshPlane, 0, navMeshYPosition, 0)
                setTranslation(extension.spec_pastureExtension.helperNavMeshPlane, 0, alignedHeight, 0)
            else
                setTranslation(extension.spec_pastureExtension.helperNavMeshPlane, 0, 0, 0)
            end
        end
    end

    self:resetBrushHeight()
end

function ConstructionActions:resetBrushHeight()
    if self.brush ~= nil and self.brush.placeable ~= nil then
        if self.brush.placeable.spec_husbandryAnimals ~= nil and self.brush.placeable.spec_husbandryAnimals.helperNavMeshPlane ~= nil then
            setTranslation(self.brush.placeable.spec_husbandryAnimals.helperNavMeshPlane, 0, 0, 0)
        elseif self.brush.placeable.spec_pastureExtension ~= nil and self.brush.placeable.spec_pastureExtension.helperNavMeshPlane ~= nil then
            setTranslation(self.brush.placeable.spec_pastureExtension.helperNavMeshPlane, 0, 0, 0)
        end
    end
end

g_ep_constructionActions = ConstructionActions.new()