local modDirectory = g_currentModDirectory

PlaceableHusbandryAnimalsExtension = {
    NAV_MESH_CLUSTER_STEP = 5
}

function PlaceableHusbandryAnimalsExtension.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(PlaceableHusbandryAnimals, specializations)
end

function PlaceableHusbandryAnimalsExtension.registerFunctions(placeableType)
	SpecializationUtil.registerFunction(placeableType, "createPastureNavMesh", PlaceableHusbandryAnimalsExtension.createPastureNavMesh)
	SpecializationUtil.registerFunction(placeableType, "createNavMeshPlanes", PlaceableHusbandryAnimalsExtension.createNavMeshPlanes)
	SpecializationUtil.registerFunction(placeableType, "reloadNavMesh", PlaceableHusbandryAnimalsExtension.reloadNavMesh)
	SpecializationUtil.registerFunction(placeableType, "updateClusterHusbandry", PlaceableHusbandryAnimalsExtension.updateClusterHusbandry)
	SpecializationUtil.registerFunction(placeableType, "generateNavMesh", PlaceableHusbandryAnimalsExtension.generateNavMesh)
	SpecializationUtil.registerFunction(placeableType, "addAnimalGrazingIfNeeded", PlaceableHusbandryAnimalsExtension.addAnimalGrazingIfNeeded)
	SpecializationUtil.registerFunction(placeableType, "addAnimalGrazingAreaIfNeeded", PlaceableHusbandryAnimalsExtension.addAnimalGrazingAreaIfNeeded)
	SpecializationUtil.registerFunction(placeableType, "removeAnimalGrazingAreaIfNeeded", PlaceableHusbandryAnimalsExtension.removeAnimalGrazingAreaIfNeeded)
	SpecializationUtil.registerFunction(placeableType, "updateNavMeshMaterial", PlaceableHusbandryAnimalsExtension.updateNavMeshMaterial)
	SpecializationUtil.registerFunction(placeableType, "navMeshMaterialLoaded", PlaceableHusbandryAnimalsExtension.navMeshMaterialLoaded)
	SpecializationUtil.registerFunction(placeableType, "loadNavMeshPlaneFile", PlaceableHusbandryAnimalsExtension.loadNavMeshPlaneFile)
	SpecializationUtil.registerFunction(placeableType, "navMeshPlaneFileLoaded", PlaceableHusbandryAnimalsExtension.navMeshPlaneFileLoaded)
	SpecializationUtil.registerFunction(placeableType, "addExtension", PlaceableHusbandryAnimalsExtension.addExtension)
	SpecializationUtil.registerFunction(placeableType, "removeExtension", PlaceableHusbandryAnimalsExtension.removeExtension)
end

function PlaceableHusbandryAnimalsExtension.registerOverwrittenFunctions(placeableType)
	SpecializationUtil.registerOverwrittenFunction(placeableType, "updateInfo", PlaceableHusbandryAnimalsExtension.updateInfo)
	SpecializationUtil.registerOverwrittenFunction(placeableType, "canBeSold", PlaceableHusbandryAnimalsExtension.canBeSold)
end

function PlaceableHusbandryAnimalsExtension.registerEventListeners(placeableType)
	SpecializationUtil.registerEventListener(placeableType, "onLoad", PlaceableHusbandryAnimalsExtension)
	SpecializationUtil.registerEventListener(placeableType, "onDelete", PlaceableHusbandryAnimalsExtension)
	SpecializationUtil.registerEventListener(placeableType, "onFinalizePlacement", PlaceableHusbandryAnimalsExtension)
	SpecializationUtil.registerEventListener(placeableType, "onReadStream", PlaceableHusbandryAnimalsExtension)
	SpecializationUtil.registerEventListener(placeableType, "onWriteStream", PlaceableHusbandryAnimalsExtension)
end

function PlaceableHusbandryAnimalsExtension.registerXMLPaths(schema, basePath)
    schema:setXMLSpecializationType("Husbandry")

	schema:register(XMLValueType.NODE_INDEX, "placeable.animalGrazingAreas.animalGrazingArea(?)#startNode", "Start node")
	schema:register(XMLValueType.NODE_INDEX, "placeable.animalGrazingAreas.animalGrazingArea(?)#widthNode", "Width node")
	schema:register(XMLValueType.NODE_INDEX, "placeable.animalGrazingAreas.animalGrazingArea(?)#heightNode", "Height node")

	basePath = basePath .. ".husbandry.animals"
	schema:register(XMLValueType.NODE_INDEX, basePath .. ".navigation#navMeshPlaneNode", "Navigation mesh plane node")
	schema:register(XMLValueType.STRING, basePath .. ".navigation#navMeshPlaneFile", "Navigation mesh plane node file")
	schema:register(XMLValueType.BOOL, basePath .. ".navigation#keepNavMeshMaterial", "Indicator to replace the navMesh material")

    schema:setXMLSpecializationType()
end

function PlaceableHusbandryAnimalsExtension.registerSavegameXMLPaths(schema, basePath)
	schema:setXMLSpecializationType("Husbandry")
	schema:register(XMLValueType.STRING, basePath .. "#pastureSystemId", "Pasture system id")
	schema:setXMLSpecializationType()
end

function PlaceableHusbandryAnimalsExtension:onLoad(savegame)
	local spec = self.spec_husbandryAnimals
    spec.animalGrazingAreas = {}

    self.xmlFile:iterate("placeable.animalGrazingAreas.animalGrazingArea", function (_, key)
		local area = {
            start = self.xmlFile:getValue(key .. "#startNode", nil, self.components, self.i3dMappings),
            width = self.xmlFile:getValue(key .. "#widthNode", nil, self.components, self.i3dMappings),
            height = self.xmlFile:getValue(key .. "#heightNode", nil, self.components, self.i3dMappings)
        }

		table.insert(spec.animalGrazingAreas, area)
	end)

    spec.info.pastureSystemId = {
		text = "",
		title = g_i18n:getText("EP_pastureSystemInfo")
	}

	spec.keepNavMeshMaterial = self.xmlFile:getValue("placeable.husbandry.animals.navigation#keepNavMeshMaterial", false)
	spec.navMeshPlaneFile = self.xmlFile:getValue("placeable.husbandry.animals.navigation#navMeshPlaneFile")

    if spec.navMeshPlaneFile then
        self:loadNavMeshPlaneFile()
    else
        spec.rootNavMeshPlane = self.xmlFile:getValue("placeable.husbandry.animals.navigation#navMeshPlaneNode", nil, self.components, self.i3dMappings)

        spec.helperNavMeshPlane = clone(spec.rootNavMeshPlane, true, false, false)
        setVisibility(spec.helperNavMeshPlane, g_ep_constructionActions.shouldShowPastureExtensionHelper or false)
        local x, y, z = getTranslation(spec.helperNavMeshPlane)
        setTranslation(spec.helperNavMeshPlane, x, y + g_ep_constructionActions.currentHeight, z)

        if not spec.keepNavMeshMaterial then
            self:updateNavMeshMaterial()
        end

        self:generateNavMesh()
    end
end

function PlaceableHusbandryAnimalsExtension:onDelete()
    local pastureSystem = g_currentMission.pastureSystem
    pastureSystem:removeExtendablePasture(self)
    local spec = self.spec_husbandryAnimals
    if spec.navMeshMaterialLoadRequestId ~= nil then
		g_i3DManager:releaseSharedI3DFile(spec.navMeshMaterialLoadRequestId)
	end
    if spec.navMeshPlaneFileLoadRequestId ~= nil then
		g_i3DManager:releaseSharedI3DFile(spec.navMeshPlaneFileLoadRequestId)
	end
end

function PlaceableHusbandryAnimalsExtension:onFinalizePlacement()
    if self.isServer then
        local pastureSystem = g_currentMission.pastureSystem
        pastureSystem:addExtendablePasture(self)
    end

    self:addAnimalGrazingIfNeeded()
end

function PlaceableHusbandryAnimalsExtension:updateInfo(superFunc, infoTable)
	superFunc(self, infoTable)

	local spec = self.spec_husbandryAnimals
    spec.info.pastureSystemId.text = self.pastureSystemId

	table.insert(infoTable, spec.info.pastureSystemId)
end

function PlaceableHusbandryAnimalsExtension:canBeSold(superFunc)
	if #g_currentMission.pastureSystem:getPastureExtensionsForPasture(self.pastureSystemId) > 0 then
		return false, g_i18n:getText("EP_husbandryAnimalsExtensionCanNotBeSold")
	end

	return superFunc(self)
end

function PlaceableHusbandryAnimalsExtension:loadNavMeshPlaneFile()
    local spec = self.spec_husbandryAnimals
    local navMeshPlaneFilename = Utils.getFilename(spec.navMeshPlaneFile, modDirectory)
    local loadingTask = self:createLoadingTask(spec)
    local arguments = {
        loadingTask = loadingTask
    }
    spec.navMeshPlaneFileLoadRequestId = g_i3DManager:loadSharedI3DFileAsync(navMeshPlaneFilename, true, false, self.navMeshPlaneFileLoaded, self, arguments)
end

function PlaceableHusbandryAnimalsExtension:navMeshPlaneFileLoaded(node, failedReason, args)
    if node == nil then
        Logging.error("Could not load navMeshMaterial")
        return
    end

    local spec = self.spec_husbandryAnimals
    spec.rootNavMeshPlane = I3DUtil.indexToObject(node, "0")
    link(spec.navigationMeshRootNode, spec.rootNavMeshPlane)

    spec.helperNavMeshPlane = clone(spec.rootNavMeshPlane, true, false, false)
    setVisibility(spec.helperNavMeshPlane, g_ep_constructionActions.shouldShowPastureExtensionHelper or false)
    local x, y, z = getTranslation(spec.helperNavMeshPlane)
    setTranslation(spec.helperNavMeshPlane, x, y + g_ep_constructionActions.currentHeight, z)

    self:generateNavMesh()

	local loadingTask = args.loadingTask
	delete(node)
	self:finishLoadingTask(loadingTask)
end

function PlaceableHusbandryAnimalsExtension:updateNavMeshMaterial()
    local spec = self.spec_husbandryAnimals
    local navMeshMaterialFilename = Utils.getFilename("materials/navMeshMaterial.i3d", modDirectory)
    local loadingTask = self:createLoadingTask(spec)
    local arguments = {
        loadingTask = loadingTask
    }
    spec.navMeshMaterialLoadRequestId = g_i3DManager:loadSharedI3DFileAsync(navMeshMaterialFilename, true, false, self.navMeshMaterialLoaded, self, arguments)
end

function PlaceableHusbandryAnimalsExtension:navMeshMaterialLoaded(node, failedReason, args)
    if node == nil then
        Logging.error("Could not load navMeshMaterial")
        return
    end

    local spec = self.spec_husbandryAnimals
    local navMeshMaterialId = I3DUtil.indexToObject(node, "0")
    local navMeshMaterial = getMaterial(navMeshMaterialId, 0)
    setMaterial(spec.helperNavMeshPlane, navMeshMaterial, 0)
    setIsNonRenderable(spec.helperNavMeshPlane, false)

	local loadingTask = args.loadingTask
	delete(node)
	self:finishLoadingTask(loadingTask)
end

function PlaceableHusbandryAnimalsExtension:addAnimalGrazingIfNeeded()
    if self.animalGrazing ~= nil and self.animalGrazing.areas ~= nil then
        for i,v in ipairs(self.spec_husbandryAnimals.animalGrazingAreas) do
            table.insert(self.animalGrazing.areas, v)
        end
    end
end

function PlaceableHusbandryAnimalsExtension:generateNavMesh(isReload)
    local spec = self.spec_husbandryAnimals

    if spec.rootNavMeshPlane == nil then
		Logging.error("Could not load root navigation mesh plane")
        return
    end

    local rootNavMeshClone = clone(spec.rootNavMeshPlane, false, false, false)

    if spec.navigationMesh ~= nil then
        delete(spec.navigationMesh)
    end

    spec.navigationMesh = self:createPastureNavMesh()
    local transformGroup = self:createNavMeshPlanes(rootNavMeshClone)

    local cellSize = 0.15
    local cellHeight = 0.2
    local agentHeight = 2.0
    local agentRadius = 0.3
    local agentMaxClimb = 0.9
    local agentMaxSlope = 45.0
    local minRegionSize = 5.0
    local mergeRegionSize = 5.0
    local maxEdgeLength = 12.0
    local maxSimplificationError = 1.5
    local navMeshBuildMask = 254
    local terrainDetail = 100
    local terrainCullInfoLayer = "infoLayer"
    local terrainCullInfoLayerChannels = 1

    local success = buildNavMesh(
        spec.navigationMesh,
        transformGroup,
        cellSize,
        cellHeight,
        agentHeight,
        agentRadius,
        agentMaxClimb,
        agentMaxSlope,
        minRegionSize,
        mergeRegionSize,
        maxEdgeLength,
        maxSimplificationError,
        navMeshBuildMask,
        terrainDetail,
        terrainCullInfoLayer,
        terrainCullInfoLayerChannels
    )

    delete(transformGroup)

    if success == false then
		Logging.error("Could not create navMesh")
    end

    if isReload then
        self:updateClusterHusbandry()
    end
end

function PlaceableHusbandryAnimalsExtension:createPastureNavMesh()
    local spec = self.spec_husbandryAnimals
    local navMesh = createNavMesh("navMesh")
    local rootX, rootY, rootZ = getWorldTranslation(spec.navigationMeshRootNode)
    local rootRotX, rootRotY, rootRotZ = getRotation(spec.navigationMeshRootNode)
    link(spec.navigationMeshRootNode, navMesh)
    setWorldTranslation(navMesh, rootX, rootY, rootZ)
    setRotation(navMesh, rootRotX, rootRotY, rootRotZ)
    return navMesh
end

function PlaceableHusbandryAnimalsExtension:createNavMeshPlanes(rootNavMeshPlane)
    local spec = self.spec_husbandryAnimals
    local pastureSystem = g_currentMission.pastureSystem
    local transformGroup = createTransformGroup("navMashPlaneGroup")

    local localX, localY, localZ = getTranslation(spec.navigationMeshRootNode)
    local rootNavMeshX, rootNavMeshY, rootNavMeshZ = localToWorld(spec.navigationMeshRootNode, localX, localY, localZ)
    local rootNavMeshRotX, rootNavMeshRotY, rootNavMeshRotZ = getWorldRotation(spec.navigationMeshRootNode)

    link(transformGroup, rootNavMeshPlane)

    setWorldTranslation(rootNavMeshPlane, rootNavMeshX, rootNavMeshY, rootNavMeshZ)
    setWorldRotation(rootNavMeshPlane, rootNavMeshRotX, rootNavMeshRotY, rootNavMeshRotZ)

    for k,v in pairs(pastureSystem:getPastureExtensionsForPasture(self.pastureSystemId)) do
        local x, y, z = getWorldTranslation(v.spec_pastureExtension.navMeshPlane)
        local rx, ry, rz = getWorldRotation(v.spec_pastureExtension.navMeshPlane)

        local navMeshClone = clone(v.spec_pastureExtension.navMeshPlane, false, false, false)
        setUserAttribute(navMeshClone, "animalType", "String", spec.animalType.name)
        link(transformGroup, navMeshClone)

        local diff = math.abs(math.abs(y) - math.abs(rootNavMeshY))
        local multiple = math.floor(diff / PlaceableHusbandryAnimalsExtension.NAV_MESH_CLUSTER_STEP)
        local addition = multiple * PlaceableHusbandryAnimalsExtension.NAV_MESH_CLUSTER_STEP

        local newYValue = rootNavMeshY
        if y > rootNavMeshY then
            newYValue = newYValue + addition
        else
            newYValue = newYValue - addition
        end
        v.spec_pastureExtension.navMeshYPosition = newYValue
        setWorldTranslation(navMeshClone, x, newYValue, z)
        setWorldRotation(navMeshClone, rx, ry, rz)
    end

    return transformGroup
end

function PlaceableHusbandryAnimalsExtension:addExtension(pastureExtension, mapLoadingFinished)
    local spec = self.spec_husbandryAnimals
    local animalTypeName = self.spec_husbandryAnimals.animalType.name
    local numAnimalsToAdd = pastureExtension:numAnimalsToAddForType(animalTypeName)
    spec.maxNumAnimals = spec.maxNumAnimals + numAnimalsToAdd
    spec.maxNumVisualAnimals = math.min(spec.maxNumAnimals, 25)
    self:addAnimalGrazingAreaIfNeeded(pastureExtension)
    if mapLoadingFinished then
        self:reloadNavMesh()
    end
end

function PlaceableHusbandryAnimalsExtension:removeExtension(pastureExtension, mapLoadingFinished)
    local spec = self.spec_husbandryAnimals
    local animalTypeName = self.spec_husbandryAnimals.animalType.name
    local numAnimalsToRemove = pastureExtension:numAnimalsToAddForType(animalTypeName)
    spec.maxNumAnimals = spec.maxNumAnimals - numAnimalsToRemove
    spec.maxNumVisualAnimals = math.min(spec.maxNumAnimals, 25)
    self:removeAnimalGrazingAreaIfNeeded(pastureExtension)
    if mapLoadingFinished then
        self:reloadNavMesh()
    end
end

function PlaceableHusbandryAnimalsExtension:addAnimalGrazingAreaIfNeeded(pastureExtension)
    if self.animalGrazing ~= nil and self.animalGrazing.areas ~= nil then
        for i,v in ipairs(pastureExtension.spec_pastureExtension.animalGrazingAreas) do
            table.insert(self.animalGrazing.areas, v)
        end
    end
end

function PlaceableHusbandryAnimalsExtension:removeAnimalGrazingAreaIfNeeded(pastureExtension)
    if self.animalGrazing ~= nil and self.animalGrazing.areas ~= nil then
        local allAreas = self.animalGrazing.areas

        for i,v in ipairs(pastureExtension.spec_pastureExtension.animalGrazingAreas) do
            for index, area in ipairs(allAreas) do
                if v.start == area.start and v.width == area.width and v.height == area.height then
                    table.remove(self.animalGrazing.areas, index)
                end
            end
        end
    end
end

function PlaceableHusbandryAnimalsExtension:reloadNavMesh()
    if self.pastureSystemId ~= nil then
        self:generateNavMesh(true)
    end
end

function PlaceableHusbandryAnimalsExtension:updateClusterHusbandry()
    local spec = self.spec_husbandryAnimals
    g_currentMission.husbandrySystem:removeClusterHusbandry(spec.clusterHusbandry)
    spec.clusterHusbandry:delete()
    spec.clusterHusbandry = nil
    spec.clusterHusbandry = AnimalClusterHusbandry.new(self, spec.animalType.name, spec.maxNumVisualAnimals)
    self:createHusbandry()
    self:updateVisualAnimals()
end

function PlaceableHusbandryAnimalsExtension:onReadStream(streamId, connection)
	self.pastureSystemId = streamReadString(streamId)

    local pastureSystem = g_currentMission.pastureSystem
    pastureSystem:addExtendablePasture(self)
end

function PlaceableHusbandryAnimalsExtension:onWriteStream(streamId, connection)
    streamWriteString(streamId, self.pastureSystemId)
end

function PlaceableHusbandryAnimalsExtension:saveToXMLFile(xmlFile, key, usedModNames)
    xmlFile:setString(key .. "#pastureSystemId", self.pastureSystemId)
end

function PlaceableHusbandryAnimalsExtension:loadFromXMLFile(xmlFile, key)
	self.pastureSystemId = xmlFile:getString(key .. "#pastureSystemId")
end