PlaceablePastureExtension = {}

function PlaceablePastureExtension.prerequisitesPresent(specializations)
	return true
end

function PlaceablePastureExtension.registerEventListeners(placeableType)
	SpecializationUtil.registerEventListener(placeableType, "onLoad", PlaceablePastureExtension)
	SpecializationUtil.registerEventListener(placeableType, "onFinalizePlacement", PlaceablePastureExtension)
	SpecializationUtil.registerEventListener(placeableType, "onDelete", PlaceablePastureExtension)
	SpecializationUtil.registerEventListener(placeableType, "onReadStream", PlaceablePastureExtension)
	SpecializationUtil.registerEventListener(placeableType, "onWriteStream", PlaceablePastureExtension)
	SpecializationUtil.registerEventListener(placeableType, "onBuy", PlaceablePastureExtension)
end

function PlaceablePastureExtension.registerFunctions(placeableType)
	SpecializationUtil.registerFunction(placeableType, "onPastureSelected", PlaceablePastureExtension.onPastureSelected)
	SpecializationUtil.registerFunction(placeableType, "numAnimalsToAddForType", PlaceablePastureExtension.numAnimalsToAddForType)
end

function PlaceablePastureExtension.registerOverwrittenFunctions(placeableType)
	SpecializationUtil.registerOverwrittenFunction(placeableType, "canBeSold", PlaceablePastureExtension.canBeSold)
	SpecializationUtil.registerOverwrittenFunction(placeableType, "updateInfo", PlaceablePastureExtension.updateInfo)
end


function PlaceablePastureExtension.registerXMLPaths(schema, basePath)
    schema:setXMLSpecializationType("PastureExtension")

    ObjectChangeUtil.registerObjectChangesXMLPaths(schema, "placeable.base")

	schema:register(XMLValueType.NODE_INDEX, "placeable.animalGrazingAreas.animalGrazingArea(?)#startNode", "Start node")
	schema:register(XMLValueType.NODE_INDEX, "placeable.animalGrazingAreas.animalGrazingArea(?)#widthNode", "Width node")
	schema:register(XMLValueType.NODE_INDEX, "placeable.animalGrazingAreas.animalGrazingArea(?)#heightNode", "Height node")

	basePath = basePath .. ".navMesh"
	schema:register(XMLValueType.NODE_INDEX, basePath .. "#planeNode", "Node for navMesh")
	schema:register(XMLValueType.NODE_INDEX, basePath .. "#root", "Root node for navMesh")

	schema:register(XMLValueType.INT, "placeable.pastureExtension#defaultNumAnimals", "Default number of animals to add")
	schema:register(XMLValueType.STRING, "placeable.pastureExtension.extension(?)#type", "Animal type")
	schema:register(XMLValueType.INT, "placeable.pastureExtension.extension(?)#numAnimals", "Number of animals to add")

    schema:setXMLSpecializationType()
end

function PlaceablePastureExtension.registerSavegameXMLPaths(schema, basePath)
	schema:setXMLSpecializationType("PastureExtension")
	schema:register(XMLValueType.STRING, basePath .. "#parentPasture", "Parent pasture system id")
	schema:setXMLSpecializationType()
end

function PlaceablePastureExtension.initSpecialization()
	g_storeManager:addSpecType("ep_numberCowExtension", "shopListAttributeIconCapacity", PlaceablePastureExtension.loadSpecValueNumberCowExtension, PlaceablePastureExtension.getSpecValueNumberCowExtension, "placeable")
	g_storeManager:addSpecType("ep_numberHorseExtension", "shopListAttributeIconCapacity", PlaceablePastureExtension.loadSpecValueNumberHorseExtension, PlaceablePastureExtension.getSpecValueNumberHorseExtension, "placeable")
	g_storeManager:addSpecType("ep_numberPigExtension", "shopListAttributeIconCapacity", PlaceablePastureExtension.loadSpecValueNumberPigExtension, PlaceablePastureExtension.getSpecValueNumberPigExtension, "placeable")
	g_storeManager:addSpecType("ep_numberSheepExtension", "shopListAttributeIconCapacity", PlaceablePastureExtension.loadSpecValueNumberSheepExtension, PlaceablePastureExtension.getSpecValueNumberSheepExtension, "placeable")
	g_storeManager:addSpecType("ep_numberChickenExtension", "shopListAttributeIconCapacity", PlaceablePastureExtension.loadSpecValueNumberChickenExtension, PlaceablePastureExtension.getSpecValueNumberChickenExtension, "placeable")
end

function PlaceablePastureExtension:onLoad(savegame)
    local spec = {}
    self.spec_pastureExtension = spec
    local xmlFile = self.xmlFile

    local objectChanges = {}

	ObjectChangeUtil.loadObjectChangeFromXML(self.xmlFile, "placeable.base.objectChanges", objectChanges, self.components, self)
	ObjectChangeUtil.setObjectChanges(objectChanges, true)

    spec.animalGrazingAreas = {}

    spec.defaultNumAnimalsToAdd = xmlFile:getInt("placeable.pastureExtension#defaultNumAnimals", 5)
    spec.numAnimalsToAdd = {}
    self.xmlFile:iterate("placeable.pastureExtension.extension", function (_, key)
        local animalType = self.xmlFile:getString(key .. "#type", nil)
        local numAnimals = self.xmlFile:getInt(key .. "#numAnimals", 5)

        if animalType ~= nil then
            spec.numAnimalsToAdd[animalType] = numAnimals
        end
	end)

    self.xmlFile:iterate("placeable.animalGrazingAreas.animalGrazingArea", function (_, key)
		local area = {
            start = self.xmlFile:getValue(key .. "#startNode", nil, self.components, self.i3dMappings),
            width = self.xmlFile:getValue(key .. "#widthNode", nil, self.components, self.i3dMappings),
            height = self.xmlFile:getValue(key .. "#heightNode", nil, self.components, self.i3dMappings)
        }

		table.insert(spec.animalGrazingAreas, area)
	end)

	spec.root = xmlFile:getValue("placeable.navMesh#root", nil, self.components, self.i3dMappings)
	spec.navMeshPlane = xmlFile:getValue("placeable.navMesh#planeNode", nil, self.components, self.i3dMappings)

    spec.helperNavMeshPlane = clone(spec.navMeshPlane, true, false, false)

    setVisibility(spec.helperNavMeshPlane, g_ep_constructionActions.shouldShowPastureExtensionHelper or false)
    local x, y, z = getTranslation(spec.helperNavMeshPlane)
    setTranslation(spec.helperNavMeshPlane, x, y + g_ep_constructionActions.currentHeight, z)
end

function PlaceablePastureExtension:onDelete()
    local pastureSystem = g_currentMission.pastureSystem
    pastureSystem:removePastureExtension(self)
end

function PlaceablePastureExtension:onFinalizePlacement()
    local pastureSystem = g_currentMission.pastureSystem
    pastureSystem:addPastureExtension(self, self.parentPasture, true)
end

function PlaceablePastureExtension:onBuy()
    if not g_currentMission.missionDynamicInfo.isMultiplayer then
        g_pasturePickerDialog:setCallback(self.onPastureSelected, self)
        g_pasturePickerDialog:setCurrentParentPasture(self.parentPasture)
        g_gui:showDialog("EP_PasturePickerDialog")
    end
end

function PlaceablePastureExtension:canBeSold(superFunc)
	local parent = g_currentMission.pastureSystem:pastureWithId(self.parentPasture)

    if parent ~= nil then
        local animalTypeName = parent.spec_husbandryAnimals.animalType.name
        local numAnimalsAfterSale = parent.spec_husbandryAnimals:getMaxNumOfAnimals() - self:numAnimalsToAddForType(animalTypeName)
        if parent.spec_husbandryAnimals:getNumOfAnimals() > numAnimalsAfterSale then
            return false, g_i18n:getText("EP_placeablePastureExtensionCanNotBeSold")
        else
            return superFunc(self)
        end
    else
        return superFunc(self)
    end
end

function PlaceablePastureExtension:updateInfo(superFunc, infoTable)
    superFunc(self, infoTable)
    local parent = g_currentMission.pastureSystem:pastureWithId(self.parentPasture)
    if parent ~= nil then
        parent:updateInfo(infoTable)
    end
end

function PlaceablePastureExtension:onPastureSelected(success, pastureSystemId)
    if success then
        local pastureSystem = g_currentMission.pastureSystem
        pastureSystem:addPastureExtension(self, pastureSystemId)

        if not self.isServer then
            g_client:getServerConnection():sendEvent(EP_AddPastureExtension.new(self, pastureSystemId))
        end
    end
end

function PlaceablePastureExtension:numAnimalsToAddForType(animalType)
    local spec = self.spec_pastureExtension
    local numAnimalsToAdd = spec.numAnimalsToAdd[animalType] or spec.defaultNumAnimalsToAdd
    return numAnimalsToAdd
end

function PlaceablePastureExtension.loadSpecValueNumberAnimalExtension(xmlFile, animalType)
    local data = nil

	if xmlFile:hasProperty("placeable.pastureExtension.extension") then
        xmlFile:iterate("placeable.pastureExtension.extension", function (_, key)
            local animalTypeName = xmlFile:getString(key .. "#type", nil)
            local numAnimals = xmlFile:getInt(key .. "#numAnimals", 5)
            local animalType = g_currentMission.animalSystem:getTypeByIndex(animalType)
            if animalType.name == animalTypeName then
                data = {
                    maxNumAnimals = numAnimals,
                    animalTypeName = animalType.name
                }
            end
        end)
		
	end

	return data
end

function PlaceablePastureExtension.loadSpecValueNumberCowExtension(xmlFile, customEnvironment, baseDir)
    return PlaceablePastureExtension.loadSpecValueNumberAnimalExtension(xmlFile, AnimalType.COW)
end

function PlaceablePastureExtension.loadSpecValueNumberHorseExtension(xmlFile, customEnvironment, baseDir)
    return PlaceablePastureExtension.loadSpecValueNumberAnimalExtension(xmlFile, AnimalType.HORSE)
end

function PlaceablePastureExtension.loadSpecValueNumberPigExtension(xmlFile, customEnvironment, baseDir)
    return PlaceablePastureExtension.loadSpecValueNumberAnimalExtension(xmlFile, AnimalType.PIG)
end

function PlaceablePastureExtension.loadSpecValueNumberSheepExtension(xmlFile, customEnvironment, baseDir)
    return PlaceablePastureExtension.loadSpecValueNumberAnimalExtension(xmlFile, AnimalType.SHEEP)
end

function PlaceablePastureExtension.loadSpecValueNumberChickenExtension(xmlFile, customEnvironment, baseDir)
    return PlaceablePastureExtension.loadSpecValueNumberAnimalExtension(xmlFile, AnimalType.CHICKEN)
end

function PlaceablePastureExtension.getSpecValueNumberAnimalExtension(storeItem, animalTypeIndex)
    local data = nil
    local profile = nil

    if animalTypeIndex == AnimalType.COW then
		data = storeItem.specs.ep_numberCowExtension
		profile = "shopListAttributeIconCow"
	elseif animalTypeIndex == AnimalType.SHEEP then
		data = storeItem.specs.ep_numberSheepExtension
		profile = "shopListAttributeIconSheep"
	elseif animalTypeIndex == AnimalType.HORSE then
		data = storeItem.specs.ep_numberHorseExtension
		profile = "shopListAttributeIconHorse"
	elseif animalTypeIndex == AnimalType.PIG then
		data = storeItem.specs.ep_numberPigExtension
		profile = "shopListAttributeIconPig"
	elseif animalTypeIndex == AnimalType.CHICKEN then
		data = storeItem.specs.ep_numberChickenExtension
		profile = "shopListAttributeIconChicken"
	end

    if data == nil then
        return nil
    end

    return data.maxNumAnimals, profile
end

function PlaceablePastureExtension.getSpecValueNumberCowExtension(storeItem, realItem)
    return PlaceablePastureExtension.getSpecValueNumberAnimalExtension(storeItem, AnimalType.COW)
end

function PlaceablePastureExtension.getSpecValueNumberHorseExtension(storeItem, realItem)
    return PlaceablePastureExtension.getSpecValueNumberAnimalExtension(storeItem, AnimalType.HORSE)
end

function PlaceablePastureExtension.getSpecValueNumberPigExtension(storeItem, realItem)
    return PlaceablePastureExtension.getSpecValueNumberAnimalExtension(storeItem, AnimalType.PIG)
end

function PlaceablePastureExtension.getSpecValueNumberSheepExtension(storeItem, realItem)
    return PlaceablePastureExtension.getSpecValueNumberAnimalExtension(storeItem, AnimalType.SHEEP)
end

function PlaceablePastureExtension.getSpecValueNumberChickenExtension(storeItem, realItem)
    return PlaceablePastureExtension.getSpecValueNumberAnimalExtension(storeItem, AnimalType.CHICKEN)
end

function PlaceablePastureExtension:onReadStream(streamId, connection)
    local hasParentPasture = streamReadBool(streamId)
    if hasParentPasture then
        local pastureSystem = g_currentMission.pastureSystem
        pastureSystem:removePastureExtension(self)
        self.parentPasture = streamReadString(streamId)
        pastureSystem:addPastureExtension(self, self.parentPasture)
    end
end

function PlaceablePastureExtension:onWriteStream(streamId, connection)
    local hasParentPasture = self.parentPasture ~= nil
    streamWriteBool(streamId, hasParentPasture)
    if hasParentPasture then
        streamWriteString(streamId, self.parentPasture)
    end
end

function PlaceablePastureExtension:saveToXMLFile(xmlFile, key, usedModNames)
    if self.parentPasture ~= nil then
        xmlFile:setString(key .. "#parentPasture", self.parentPasture)
    end
end

function PlaceablePastureExtension:loadFromXMLFile(xmlFile, key)
	self.parentPasture = xmlFile:getString(key .. "#parentPasture")
end