EP_PasturePickerDialog = {
	CONTROLS = {
		DIALOG_TITLE = "dialogTitleElement",
		DIALOG_SUBTITLE = "dialogSubtitleElement",
        SELECTED_PASTURE = "selectedPastureElement",
		OK_BUTTON = "okButton",
		CANCEL_BUTTON = "cancelButton"
	}
}
local EP_PasturePickerDialog_mt = Class(EP_PasturePickerDialog, MessageDialog)

function EP_PasturePickerDialog.new(target, custom_mt)
	local self = DialogElement.new(target, custom_mt or EP_PasturePickerDialog_mt)
	self.inputDelay = 250
    self.currentIndex = 1

    self:resetUI()

	self:registerControls(EP_PasturePickerDialog.CONTROLS)

	return self
end

function EP_PasturePickerDialog:resetUI()
    self.availablePastures = {}
end

function EP_PasturePickerDialog:onOpen()
	EP_PasturePickerDialog:superClass().onOpen(self)

    self:resetUI()

    local currentPasture = g_currentMission.pastureSystem:pastureWithId(self.currentParentPastureId)
    if currentPasture ~= nil then
        self.dialogSubtitleElement:setText(string.format(g_i18n:getText("EP_pasturePickerSelectedPastureSubtitle"), currentPasture:getName(), self.currentParentPastureId))
    else
        self.dialogSubtitleElement:setText("")
    end

    self.availablePastures = g_currentMission.pastureSystem:availablePasturesNameAndId(self.currentParentPastureId)
    if #self.availablePastures == 0 then
        Logging.error("No available pastures")
        return
    end

    if self.availablePastures[self.currentIndex] == nil then
        self.currentIndex = 1
    end

    local selectedPasture = self.availablePastures[self.currentIndex]

	self.inputDelay = self.time + 250

	self:updateSelectedPasture(selectedPasture)
end

function EP_PasturePickerDialog:setCallback(callbackFunc, target)
	self.callbackFunc = callbackFunc
	self.target = target
end

function EP_PasturePickerDialog:setCurrentParentPasture(currentParentPastureId)
	self.currentParentPastureId = currentParentPastureId
end

function EP_PasturePickerDialog:onClickOk()
	if self.inputDelay < self.time then
		self:sendCallback(true)

		return false
	end

	return true
end

function EP_PasturePickerDialog:onClickCancel()
	if self.inputDelay < self.time then
		self:sendCallback(false)

		return false
	end

	return true
end

function EP_PasturePickerDialog:sendCallback(success)
	if self.inputDelay < self.time then
		self:close()

		if self.callbackFunc ~= nil then
			if self.target ~= nil then
				self.callbackFunc(self.target, success, self.selectedPasture.pastureSystemId)
			else
				self.callbackFunc(success, self.selectedPasture.pastureSystemId)
			end
		end
	end
end

function EP_PasturePickerDialog:onClickLeft(element)
    if self.currentIndex == 1 then
        self.currentIndex = #self.availablePastures
    else
        self.currentIndex = self.currentIndex - 1
    end

	local pasture = self.availablePastures[self.currentIndex]

	self:updateSelectedPasture(pasture)
end

function EP_PasturePickerDialog:onClickRight(element)
	if self.currentIndex == #self.availablePastures then
        self.currentIndex = 1
    else
        self.currentIndex = self.currentIndex + 1
    end

	local pasture = self.availablePastures[self.currentIndex]

	self:updateSelectedPasture(pasture)
end

function EP_PasturePickerDialog:updateSelectedPasture(pasture)
    self.selectedPasture = pasture
	self.selectedPastureElement:setText(string.format(g_i18n:getText("EP_pasturePickerSelectedPastureText"), self.selectedPasture.pastureName, self.selectedPasture.pastureSystemId))
end