local modName = g_currentModName
local modDirectory = g_currentModDirectory

local function initPastureSystem()
    source(Utils.getFilename("src/PastureSystem.lua", g_currentModDirectory))
    source(Utils.getFilename("src/events/EP_AddPastureExtension.lua", g_currentModDirectory))

    FSBaseMission.initialize = Utils.appendedFunction(FSBaseMission.initialize, function(self)
		self.pastureSystem = PastureSystem.new()
	end)

    FSBaseMission.onFinishedLoading = Utils.appendedFunction(FSBaseMission.onFinishedLoading, function(self)
        if g_currentMission:getIsServer() then
            g_currentMission.pastureSystem.mapLoadingFinished = true
            g_currentMission.pastureSystem:reloadAllPastures()
        end
	end)

    FSBaseMission.onStartMission = Utils.appendedFunction(FSBaseMission.onStartMission, function(self)
        if not g_currentMission:getIsServer() then
            g_currentMission.pastureSystem.mapLoadingFinished = true
            g_currentMission.pastureSystem:reloadAllPastures()
        end
	end)
end

local function initGui()
    source(Utils.getFilename("src/gui/EP_PasturePickerDialog.lua", g_currentModDirectory))

    g_gui:loadProfiles(Utils.getFilename("gui/EP_GuiProfiles.xml", g_currentModDirectory))

    g_pasturePickerDialog = EP_PasturePickerDialog.new(nil, nil, g_inputBinding)
	g_gui:loadGui(Utils.getFilename("gui/EP_PasturePickerDialog.xml", g_currentModDirectory), "EP_PasturePickerDialog", g_pasturePickerDialog)
end

local function initPastureExtensionHelper()
    source(Utils.getFilename("src/ConstructionActions.lua", g_currentModDirectory))
    source(Utils.getFilename("src/ConstructionBrushPlaceableExtension.lua", g_currentModDirectory))
end

local function addConstructionCategoryTabs(self, superFunc)
    for k,v in pairs(g_storeManager.modCategoryTabs) do
        for i,tab in ipairs(v) do
            g_storeManager:addConstructionTab(k, tab.name, tab.title, tab.iconFilename, tab.iconUVs, modDirectory)
        end
    end

    return superFunc(self)
end

local function initPastureExtensionsCategory()
    StoreManager.getDefaultStoreItemsFilename = Utils.overwrittenFunction(StoreManager.getDefaultStoreItemsFilename, addConstructionCategoryTabs)

    g_storeManager.modCategoryTabs = {}
    local modDescPath = Utils.getFilename("modDesc.xml", modDirectory)
    local modDesc = XMLFile.load("modDesc", modDescPath)

    modDesc:iterate("modDesc.constructionCategoryTabs.tab", function (_, key)
        local parentCategory = modDesc:getString(key.."#parentCategory")
        local allTabs = g_storeManager.modCategoryTabs[parentCategory] or {}
        local refSize = modDesc:getVector(key.."#refSize", { 1024, 1024 }, 2)
        local tab = {
            name = modDesc:getString(key.."#name"),
            title = g_i18n:convertText(modDesc:getString(key.."#title"), modName),
            iconFilename = modDesc:getString(key.."#iconFilename"),
            iconUVs = GuiUtils.getUVs(modDesc:getString(key.."#iconUVs", "0 0 1 1"), refSize)
        }
        table.insert(allTabs, tab)
        g_storeManager.modCategoryTabs[parentCategory] = allTabs
    end)
end

initPastureSystem()
initGui()
initPastureExtensionHelper()
initPastureExtensionsCategory()